#include "lasp_cppdaq.h"
#include <algorithm>
#include <cassert>
#include "lasp_config.h"

#define MAX_DEV_COUNT_PER_API 20

#ifdef LASP_HAS_ULDAQ
#include "lasp_cppuldaq.h"
#endif
#ifdef LASP_HAS_RTAUDIO
#include "lasp_cpprtaudio.h"
#endif

vector<DeviceInfo> Daq::getDeviceInfo() {
  vector<DeviceInfo> devs;
#ifdef LASP_HAS_ULDAQ
  fillUlDaqDeviceInfo(devs);
#endif

#ifdef LASP_HAS_RTAUDIO
  fillRtAudioDeviceInfo(devs);
#endif

  return devs;
}

vector<DaqApi> DaqApi::getAvailableApis() {

  vector<DaqApi> apis;
  apis.resize(6);
#ifdef LASP_HAS_ULDAQ
  apis.at(uldaqapi.apicode) = uldaqapi;
#endif
#ifdef LASP_HAS_RTAUDIO
  apis.at(rtaudioAlsaApi.apicode) = rtaudioAlsaApi;
  apis.at(rtaudioPulseaudioApi.apicode) = rtaudioPulseaudioApi;
  apis.at(rtaudioWasapiApi.apicode) = rtaudioWasapiApi;
  apis.at(rtaudioDsApi.apicode) = rtaudioDsApi;
  apis.at(rtaudioAsioApi.apicode) = rtaudioAsioApi;
#endif
  return apis;
}


DaqConfiguration::DaqConfiguration(const DeviceInfo &device) {

  api = device.api;
  device_name = device.device_name;

  eninchannels.resize(device.ninchannels, false);
  enoutchannels.resize(device.noutchannels, false);

  inchannel_sensitivities.resize(device.ninchannels, 1.0);
  inchannel_metadata.resize(device.ninchannels, "");
  for (us i = 0; i < eninchannels.size(); i++) {
    std::stringstream chname;
    chname << "Unnamed input channel " << i;
    inchannel_names.push_back(chname.str());

  }

  outchannel_metadata.resize(device.noutchannels, "");
  outchannel_sensitivities.resize(device.noutchannels, 1.0);
  for (us i = 0; i < enoutchannels.size(); i++) {
    std::stringstream chname;
    chname << "Unnamed output channel " << i;
    outchannel_names.push_back(chname.str());
  }
  sampleRateIndex = device.prefSampleRateIndex;
  dataTypeIndex = device.prefDataTypeIndex;
  framesPerBlockIndex = device.prefFramesPerBlockIndex;

  monitorOutput = false;

  inputIEPEEnabled.resize(device.ninchannels, false);
  inputACCouplingMode.resize(device.ninchannels, false);
  inputRangeIndices.resize(device.ninchannels, device.prefInputRangeIndex);

  assert(match(device));
}

bool DaqConfiguration::match(const DeviceInfo& dev) const {
  return (dev.device_name == device_name && dev.api == api);
}

int DaqConfiguration::getHighestInChannel() const {
  for(int i=eninchannels.size()-1; i>-1;i--) {
    if(eninchannels.at(i)) return i;
  }
  return -1;
}

int DaqConfiguration::getHighestOutChannel() const {
  for(us i=enoutchannels.size()-1; i>=0;i--) {
    if(enoutchannels.at(i)) return i;
  }
  return -1;
}
int DaqConfiguration::getLowestInChannel() const {
  for(us i=0; i<eninchannels.size();i++) {
    if(eninchannels.at(i)) return i;
  }
  return -1;
}
int DaqConfiguration::getLowestOutChannel() const {
  for(us i=0; i<enoutchannels.size();i++) {
    if(enoutchannels.at(i)) return i;
  }
  return -1;
}

Daq *Daq::createDaq(const DeviceInfo& devinfo,
    const DaqConfiguration &config) {

  if(!config.match(devinfo)) {
    throw runtime_error("DaqConfiguration does not match device info");
  }

  // Some basic sanity checks
  if ((devinfo.ninchannels != config.eninchannels.size())) {
    /* cerr << "devinfo.ninchannels: " << devinfo.ninchannels << endl; */
    /* cerr << "config.eninchannels.size(): " << config.eninchannels.size() << endl; */
    throw runtime_error("Invalid length of enabled input channels specified");
  }
  if ((devinfo.noutchannels != config.enoutchannels.size())) {
    throw runtime_error("outvalid length of enabled output channels specified");
  }

  int apicode = devinfo.api.apicode;
  if(devinfo.api == DaqApi()) {
    throw std::runtime_error(string("Unable to match API: ") + devinfo.api.apiname);
  }
#ifdef LASP_HAS_ULDAQ
  else if (devinfo.api == uldaqapi) {
    return createUlDaqDevice(devinfo, config);
  }
#endif
#ifdef LASP_HAS_RTAUDIO
  else if(apicode >= 1 && apicode <= 5) {
    return createRtAudioDevice(devinfo, config);
  }
#endif
  else {
    throw std::runtime_error(string("Unable to match API: ") +
        devinfo.api.apiname);
  }
}


Daq::Daq(const DeviceInfo &devinfo, const DaqConfiguration &config)
  : DaqConfiguration(config), DeviceInfo(devinfo) {

    if (monitorOutput && !(nenoutchannels() > 0)) {
      throw runtime_error(
          "Output monitoring only possible when at least one output channel is enabled. Please make sure to enable at least one output channel");
    }
  }


double Daq::samplerate() const {
  mutexlock lock(mutex);
  assert(sampleRateIndex < availableSampleRates.size());
  return availableSampleRates.at(sampleRateIndex);
}


DataType Daq::dataType() const {
  mutexlock lock(mutex);
  assert((us)dataTypeIndex < availableDataTypes.size());
  return availableDataTypes.at(dataTypeIndex);
}


double Daq::inputRangeForChannel(us ch) const {
  if (!(ch < ninchannels)) {
    throw runtime_error("Invalid channel number");
  }
  mutexlock lock(mutex);
  assert(inputRangeIndices.size() == eninchannels.size());
  return availableInputRanges.at(inputRangeIndices.at(ch));
}


us Daq::neninchannels(bool include_monitorchannel) const {
  mutexlock lock(mutex);
  us inch = std::count(eninchannels.begin(), eninchannels.end(), true);
  if (monitorOutput && include_monitorchannel) {
    inch++;
  }
  return inch;
}


us Daq::nenoutchannels() const {
  mutexlock lock(mutex);
  return std::count(enoutchannels.begin(), enoutchannels.end(), true);
}
