cimport cython
from ..lasp_common import AvType
from .lasp_deviceinfo cimport DeviceInfo
from .lasp_daqconfig cimport DaqConfiguration
from numpy cimport import_array

from cpython.ref cimport PyObject,Py_INCREF, Py_DECREF
import numpy as np
import logging

__all__ = ['Daq']

cdef cnp.NPY_TYPES getCNumpyDataType(DataType& dt):
    if(dt == dtype_fl32):
        return cnp.NPY_FLOAT32
    elif(dt == dtype_fl64):
        return cnp.NPY_FLOAT64
    elif(dt == dtype_int8):
        return cnp.NPY_INT8
    elif(dt == dtype_int16):
        return cnp.NPY_INT16
    elif(dt == dtype_int32):
        return cnp.NPY_INT32
    else:
        raise ValueError('Unknown data type')

cdef getNumpyDataType(DataType& dt):
    if(dt == dtype_fl32):
        return np.dtype(np.float32)
    elif(dt == dtype_fl64):
        return np.dtype(np.float64)
    elif(dt == dtype_int8):
        return np.dtype(np.int8)
    elif(dt == dtype_int16):
        return np.dtype(np.int16)
    elif(dt == dtype_int32):
        return np.dtype(np.int32)
    else:
        raise ValueError('Unknown data type')

DEF QUEUE_BUFFER_TIME = 0.5

ctypedef struct PyStreamData:
    PyObject* pyCallback

    # Flag used to pass the stopThread.
    atomic[bool] stopThread

    # Flag to indicate that the signal generator queue has been filled for the
    # first time.
    atomic[bool] ready

    # Number of frames per block
    unsigned nFramesPerBlock

    # Number of bytes per channel
    unsigned int nBytesPerChan

    unsigned ninchannels
    unsigned noutchannels

    double samplerate

    cnp.NPY_TYPES npy_format

    # If either of these queue pointers are NULL, it means the stream does not have an
    # input, or output.
    SafeQueue[void*] *inQueue
    SafeQueue[void*] *outQueue
    CPPThread[void*, void (*)(void*)] *thread

cdef void audioCallbackPythonThreadFunction(void* voidsd) nogil:
    cdef:
        PyStreamData* sd = <PyStreamData*> voidsd

        double* inbuffer = NULL
        double* outbuffer = NULL

        unsigned noutchannels= sd.noutchannels
        unsigned ninchannels= sd.ninchannels
        unsigned nBytesPerChan= sd.nBytesPerChan
        unsigned nFramesPerBlock= sd.nFramesPerBlock

        double sleeptime = (<double> sd.nFramesPerBlock)/(8*sd.samplerate);
        # Sleep time in microseconds
        us sleeptime_us = <us> (sleeptime*1e6);

        us nblocks_buffer = <us> max(1, (QUEUE_BUFFER_TIME * sd.samplerate /
                                         sd.nFramesPerBlock))

    with gil:
        import_array()
        npy_format = cnp.NPY_FLOAT64
        callback = <object> sd.pyCallback
        # print(f'Number of input channels: {ninchannels}')
        # print(f'Number of out channels: {noutchannels}')
        # fprintf(stderr, 'Sleep time: %d us\n', sleeptime_us)

    if sd.outQueue:
        for i in range(nblocks_buffer):
            outbuffer = <double*> malloc(sizeof(double)*nBytesPerChan*noutchannels)
            memset(outbuffer, 0, sizeof(double)*nBytesPerChan*noutchannels)
            sd.outQueue.enqueue(<double*> outbuffer)
    sd.ready.store(True)

    while not sd.stopThread.load():
        with gil:
            if sd.outQueue:
                while sd.outQueue.size() < nblocks_buffer:
                    outbuffer = <double*> malloc(sizeof(double)*nBytesPerChan*noutchannels)

                    npy_output = <object> data_to_ndarray(
                    outbuffer,
                    nFramesPerBlock,
                    noutchannels,
                    sd.npy_format,
                    False, # Do not transfer ownership to the temporary
                           # Numpy container
                    True) # F-contiguous
                    try:
                        rval = callback(None, npy_output, nFramesPerBlock)

                    except Exception as e:
                        logging.error('exception in Cython callback for audio output: ', str(e))
                        return

                    sd.outQueue.enqueue(<double*> outbuffer)

            if sd.inQueue and not sd.inQueue.empty(): 
                # Waiting indefinitely on the queue...
                inbuffer = <double*> sd.inQueue.dequeue()
                if inbuffer == NULL:
                    logging.debug('Received empty buffer on input, stopping thread...\n')
                    return

                npy_input = <object> data_to_ndarray(
                                inbuffer,
                                nFramesPerBlock,
                                ninchannels,
                                sd.npy_format,
                                True, # Do transfer ownership
                                True) # F-contiguous is True: data is Fortran-cont.

                try:
                    rval = callback(npy_input, None, nFramesPerBlock)

                except Exception as e:
                    logging.error('exception in cython callback for audio input: ', str(e))
                    return

        CPPsleep_us(sleeptime_us);

        # Outputbuffer is free'ed by the audiothread, so should not be touched
        # here.
        outbuffer = NULL

        # Inputbuffer memory is owned by Numpy, so should not be free'ed
        inbuffer = NULL

cdef class Daq:

    def __cinit__(self, DeviceInfo pydevinfo, DaqConfiguration pydaqconfig):

        """
        Acquires a daq handle, and opens the device

        """
        cdef:
            cppDaqConfiguration* daqconfig
            cppDeviceInfo* devinfo
            vector[cppDeviceInfo] devinfos

        self.daq_device = NULL
        self.sd = NULL

        daqconfig = &(pydaqconfig.config)
        devinfo = &(pydevinfo.devinfo)

        try:
            self.daq_device = cppDaq.createDaq(devinfo[0], daqconfig[0])
        except Exception as e:
            raise
        self.nFramesPerBlock = self.daq_device.framesPerBlock()
        self.samplerate = self.daq_device.samplerate()

        if self.nFramesPerBlock > 8192 or self.nFramesPerBlock < 512:
            del self.daq_device
            raise ValueError('Invalid number of nFramesPerBlock')

    def __dealloc__(self):
        # fprintf(stderr, "UlDaq.__dealloc__\n")
        if self.sd is not NULL:
            logging.debug("UlDaq.__dealloc__: stopping stream.")
            self.stop()

        if self.daq_device is not NULL:
            del self.daq_device
            self.daq_device = NULL

    def getNumpyDataType(self):
        cdef:
            DataType dt = self.daq_device.dataType()
        return getNumpyDataType(dt)


    def isRunning(self):
        return self.sd is not NULL

    @staticmethod
    def getDeviceInfo():
        cdef:
            vector[cppDeviceInfo] devinfos = cppDaq.getDeviceInfo()
        pydevinfo = {}
        for i in range(devinfos.size()):
            d = DeviceInfo()
            d.devinfo = <cppDeviceInfo> devinfos[i]
            if d.api not in pydevinfo.keys():
                pydevinfo[d.api] = [d]
            else:
                pydevinfo[d.api].append(d)

        return pydevinfo

    @cython.nonecheck(True)
    def start(self, audiocallback):
        """
        Opens a stream with specified parameters

        Args:
            avstream: AvStream instance

        Returns: None
        """
        if self.sd is not NULL:
            assert self.daq_device is not NULL
            raise RuntimeError('Stream is already opened.')

        cdef:
            cppDaq* daq = self.daq_device
            unsigned nFramesPerBlock = self.nFramesPerBlock
            DataType dtype = self.daq_device.dataType()
            double samplerate = self.samplerate


        # All set, allocate the stream!
        self.sd = <PyStreamData*> malloc(sizeof(PyStreamData))
        if self.sd == NULL:
            del daq
            raise MemoryError('Could not allocate stream: memory error.')


        self.sd.stopThread.store(False)
        self.sd.ready.store(False)

        self.sd.inQueue = NULL
        self.sd.outQueue = NULL

        self.sd.thread = NULL
        self.sd.samplerate = <double> samplerate

        self.sd.ninchannels = daq.neninchannels(True)
        self.sd.noutchannels = daq.nenoutchannels()
        self.sd.nBytesPerChan = nFramesPerBlock*dtype.sw
        self.sd.nFramesPerBlock = nFramesPerBlock
        self.sd.npy_format = getCNumpyDataType(dtype)

        if daq.neninchannels(True) > 0:
            self.sd.inQueue = new SafeQueue[void*]()
        if daq.nenoutchannels() > 0:
            self.sd.outQueue = new SafeQueue[void*]()

        self.sd.pyCallback = <PyObject*> audiocallback

        # Increase reference count to the callback
        Py_INCREF(<object> audiocallback)

        with nogil:
            self.sd.thread = new CPPThread[void*, void (*)(void*)](audioCallbackPythonThreadFunction,
            <void*> self.sd)

            while not self.sd.ready.load():
               # Allow stream stome time to start
                CPPsleep_ms(100)

        self.daq_device.start(
            self.sd.inQueue,
            self.sd.outQueue)

        return self.daq_device.samplerate()

    def stop(self):
        if self.sd is NULL:
            raise RuntimeError('Stream is not opened')

        self.cleanupStream(self.sd)
        self.sd = NULL

    cdef cleanupStream(self, PyStreamData* sd):

        with nogil:
            if sd.thread:
                sd.stopThread.store(True)

            if sd.inQueue:
                # If waiting in the input queue, hereby we let it run.
                sd.inQueue.enqueue(NULL)

            sd.thread.join()
            del sd.thread
            sd.thread = NULL

            if sd.inQueue:
                while not sd.inQueue.empty():
                    free(sd.inQueue.dequeue())
                del sd.inQueue

            if sd.outQueue:
                while not sd.outQueue.empty():
                    free(sd.outQueue.dequeue())
                del sd.outQueue
                sd.outQueue = NULL
        logging.debug("End cleanup stream queues...\n")

        if sd.pyCallback:
            Py_DECREF(<object> sd.pyCallback)
            sd.pyCallback = NULL

        free(sd)

