#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
"""
Author: J.A. de Jong - ASCEE

Description: Signal generator code

"""
import multiprocessing as mp
import dataclasses
import logging
from typing import Tuple
import numpy as np

from .filter import PinkNoise
from .lasp_octavefilter import SosOctaveFilterBank, SosThirdOctaveFilterBank
from .filter import OctaveBankDesigner, PinkNoise, ThirdOctaveBankDesigner
from .lasp_avstream import StreamManager, ignoreSigInt
from .wrappers import Siggen as pyxSiggen, Equalizer
from enum import Enum, unique, auto

QUEUE_BUFFER_TIME = 0.5  # The amount of time used in the queues for buffering
# of data, larger is more stable, but also enlarges latency

__all__ = ["SignalType", "NoiseType", "SiggenMessage", "SiggenData", "Siggen"]


@unique
class SignalType(Enum):
    Periodic = 0
    Noise = 1
    Sweep = 2
    Meas = 3


@unique
class NoiseType(Enum):
    white = "White noise"
    pink = "Pink noise"

    def __str__(self):
        return str(self.value)

    @staticmethod
    def fillComboBox(combo):
        for type_ in list(NoiseType):
            combo.addItem(str(type_))

    @staticmethod
    def getCurrent(cb):
        return list(NoiseType)[cb.currentIndex()]


class SiggenWorkerDone(Exception):
    def __str__(self):
        return "Done generating signal"


@unique
class SiggenMessage(Enum):
    """
    Different messages that can be send to the signal generator over the pipe
    connection.
    """
    endProcess = auto()  # Stop and quit the signal generator
    adjustVolume = auto()  # Adjust the volume
    newEqSettings = auto()  # Forward new equalizer settings
    newSiggenData = auto()  # Forward new equalizer settings
    ready = auto() # Send out once, once the signal generator is ready with
    # pre-generating data.
    mute = auto() # Mute / unmute siggen

    # These messages are send back to the main thread over the pipe
    error = auto()
    done = auto()


@dataclasses.dataclass
class SiggenData:
    """
    Metadata used to create a Signal Generator
    """
    fs: float  # Sample rate [Hz]

    # Number of frames "samples" to send in one block
    nframes_per_block: int

    # The data type to output
    dtype: np.dtype

    # Muted?
    muted: bool

    # Level of output signal [dBFS]el
    level_dB: float

    # Signal type specific data, i.e.
    signaltype: SignalType
    signaltypedata: Tuple = None

    # Settings for the equalizer etc
    eqdata: object = None # Equalizer data

class SiggenProcess(mp.Process):
    """
    Main function running in a different process, is responsible for generating
    new signal data. Uses the signal queue to push new generated signal data
    on.
    """
    def __init__(self, siggendata, dataq, pipe):

        """
        Args:
            siggendata: The signal generator data to start with.
            dataq: The queue to put generated signal on
            pipe: Control and status messaging pipe
        """
        super().__init__()

        # When this is set, a kill on the main process will also kill the
        # siggen process. Highly wanted feature
        self.daemon = True

        self.dataq = dataq
        self.siggendata = siggendata
        self.pipe = pipe
        self.eq = None
        self.siggen = None

        fs = self.siggendata.fs
        nframes_per_block = siggendata.nframes_per_block
        self.nblocks_buffer = max(
            1, int(QUEUE_BUFFER_TIME * fs/ nframes_per_block)
        )

    def newSiggen(self, siggendata: SiggenData):
        """
        Create a signal generator based on parameters specified in global
        function data.

        Args:
            siggendata: SiggenData. Metadata to create a new signal generator.
        """

        logging.debug('newSiggen')

        fs = siggendata.fs
        nframes_per_block = siggendata.nframes_per_block
        level_dB = siggendata.level_dB
        signaltype = siggendata.signaltype
        signaltypedata = siggendata.signaltypedata

        # Muted state
        self.muted = siggendata.muted

        if signaltype == SignalType.Periodic:
            freq, = signaltypedata
            siggen = pyxSiggen.sineWave(fs, freq, level_dB)
        elif signaltype == SignalType.Noise:
            noisetype, zerodBpoint = signaltypedata
            if noisetype == NoiseType.white:
                sos_colorfilter = None
            elif noisetype == NoiseType.pink:
                sos_colorfilter = PinkNoise(fs, zerodBpoint).flatten()
            else:
                raise ValueError(f"Unknown noise type")

            siggen = pyxSiggen.noise(fs, level_dB, sos_colorfilter)

        elif signaltype == SignalType.Sweep:
            fl, fu, Ts, Tq, sweep_flags = signaltypedata
            siggen = pyxSiggen.sweep(fs, fl, fu, Ts, Tq, sweep_flags, level_dB)

        else:
            raise ValueError(f"Not implemented signal type: {signaltype}")

        logging.debug('newSiggen')
        return siggen

    def generate(self):
        """
        Generate a single block of data and put it on the data queue
        """
        signal = self.siggen.genSignal(self.siggendata.nframes_per_block)
        dtype = self.siggendata.dtype
        if self.eq is not None:
            signal = self.eq.equalize(signal)
        if np.issubdtype(dtype, np.integer):
            bitdepth_fixed = dtype.itemsize * 8
            signal *= 2 ** (bitdepth_fixed - 1) - 1
        if self.muted:
            # Mute it
            signal *= 0
        try:
            self.dataq.put(signal.astype(dtype))
        except ValueError:
            # As of Python 3.8, a value error on a Queue means that the oter
            # end of the process died.
            logging.error("Error with data queue, terminating process")
            self.terminate()

    def newEqualizer(self, eqdata):
        """
        Create an equalizer object from equalizer data

        Args:
            eqdata: dictionary containing equalizer data. TODO: document the
            requiring fields.
        """
        if eqdata is None:
            return None
        eq_type = eqdata['type']
        eq_levels = eqdata['levels']
        fs = self.siggendata.fs

        if eq_type == 'three':
            fb = SosThirdOctaveFilterBank(fs)
        elif eq_type == 'one':
            fb = SosOctaveFilterBank(fs)

        eq = Equalizer(fb._fb)
        if eq_levels is not None:
            eq.setLevels(eq_levels)
        return eq

    def sendPipe(self, msgtype, msg):
        try:
            self.pipe.send((msgtype, msg))
        except OSError:
            logging.error("Error with pipe, terminating process")
            self.terminate()

    def run(self):
        # The main function of the actual process
        # First things first
        ignoreSigInt()

        try:
            self.siggen = self.newSiggen(self.siggendata)
        except Exception as e:
            self.sendPipe(SiggenMessage.error, str(e))

        try:
            self.eq = self.newEqualizer(self.siggendata.eqdata)
        except Exception as e:
            self.sendPipe(SiggenMessage.error, str(e))

        # Pre-generate blocks of signal data
        while self.dataq.qsize() < self.nblocks_buffer:
            self.generate()

        self.sendPipe(SiggenMessage.ready, None)

        while True:
            # Wait here for a while, to check for messages to consume
            if self.pipe.poll(timeout=QUEUE_BUFFER_TIME / 4):
                try:
                    msg, data = self.pipe.recv()
                except OSError:
                    logging.error("Error with pipe, terminating process")
                    self.terminate()

                if msg == SiggenMessage.endProcess:
                    logging.debug("Signal generator caught 'endProcess' message. Exiting.")
                    return 0
                elif msg == SiggenMessage.mute:
                    self.muted = data
                elif msg == SiggenMessage.adjustVolume:
                    level_dB = data
                    logging.debug(f"Signal generator caught 'adjustVolume' message. New volume = {level_dB:.1f} dB FS")
                    self.siggen.setLevel(level_dB)
                elif msg == SiggenMessage.newEqSettings:
                    eqdata = data
                    self.eq = self.newEqualizer(eqdata)
                elif msg == SiggenMessage.newSiggenData:
                    siggendata = data
                    self.siggen = self.newSiggen(siggendata)
                else:
                    self.pipe.send(
                        SiggenMessage.error, "BUG: Generator caught unknown message. Quiting"
                    )
            while self.dataq.qsize() < self.nblocks_buffer:
                # Generate new data and put it in the queue!
                try:
                    self.generate()
                except SiggenWorkerDone:
                    self.sendPipe(SiggenMessage.done, None)
                    return 0

        return 1


class Siggen:
    """
    Signal generator class, generates signal data in a different process to
    unload the work in the calling thread.
    """

    def __init__(self, dataq, siggendata: SiggenData):
        """"""

        self.pipe, client_end = mp.Pipe(duplex=True)

        self.stopped = False

        self.process = SiggenProcess(siggendata, dataq, client_end)
        self.process.start()

        if not self.process.is_alive():
            raise RuntimeError('Unexpected signal generator exception')

        # Block waiting here for signal generator to be ready
        msg, data = self.pipe.recv()
        if msg == SiggenMessage.ready:
            logging.debug('Signal generator ready')
        elif msg == SiggenMessage.error:
            e = data
            raise RuntimeError(f'Signal generator exception: {str(e)}')
        else:
            # Done, or something
            if msg == SiggenMessage.done:
                self.stopped = True

        self.handle_msgs()

    def setLevel(self, new_level):
        """
        Set a new signal level to the generator

        Args:
            new_level: The new level in [dBFS]
        """
        self.pipe.send((SiggenMessage.adjustVolume, new_level))

    def mute(self, mute):
        self.pipe.send((SiggenMessage.mute, mute))

    def setEqData(self, eqdata):
        self.pipe.send((SiggenMessage.newEqSettings, eqdata))

    def setSiggenData(self, siggendata: SiggenData):
        """
        Updates the whole signal generator, based on new signal generator data.
        """
        self.pipe.send((SiggenMessage.newSiggenData, siggendata))

    def handle_msgs(self):
        while self.pipe.poll():
            msg, data = self.pipe.recv()
            if msg == SiggenMessage.error:
                raise RuntimeError(
                    f"Error in initialization of signal generator: {data}"
                )
            # elif msg == SiggenMessage.done:
            #     self.stop()


    def cleanup(self):
        logging.debug('Siggen::stop()')
        self.pipe.send((SiggenMessage.endProcess, None))
        self.pipe.close()
        
        logging.debug('Joining siggen process')
        self.process.join()
        logging.debug('Joining siggen process done')
        self.process.close()

        self.process = None

